/*

acconfig.h - template used by autoheader to create config.h.in
config.h.in - used by autoconf to create config.h
config.h - created by autoconf; contains defines generated by autoconf

*/

@TOP@

/* Is the mailspool group writable */
#undef MAILGROUP

